// Dropper v1.2
// Copyright (c) 1999-2002 by Gergely Kutenics - gem@mailbox.hu

#include "VimCon.h"

#include <new>
#include <iostream>
#include <list>
#include <set>
using namespace std;




// printf() like messages
#define PM if (ToPrint) VMSG
#define PW VMSG("Warning: "),VMSG
#define PE VMSG("Error: "),VMSG
#define PFlag if (ToPrint) DoFlag
#define PBits if (ToPrint) DoBits



//
// flags printer
//

typedef struct {
	int F;
	char *S;
} DFlag;

typedef struct {
	int F;
	char *S;
} DBits;

#define DoFlag(xxx,f) PrintDFlag(xxx,members(xxx),f)
#define DoBits(xxx,f) PrintDBits(xxx,members(xxx),f)

void PrintDFlag(DFlag *p,int n,int f)
{
	for (int i=0;i<n;i++) {
		if (f==p[i].F) {
			VMSG(p[i].S);
			break;
		}
	}
	if (i==n) VMSG("unknown!");
	VMSG(" (%d)",f);
}

void PrintDBits(DBits *p,int n,int f)
{
	VMSG("0x%08X\n",f);
	for (int i=0;i<n;i++) {
		if (f&p[i].F) {
			VMSG(" - %s\n",p[i].S);
			f^=p[i].F; // remove processed flags
		}
	}
	if (f) VMSG(" - unknown flags: 0x%08X (%d)\n",f,f);
}



//
// PE program class
//


class DProg
{
public:
	DProg();
	~DProg();

	// memory
	int Alloc(int len);
	int Realloc(int len);
	int Set(byte *buf,int len);
	int Insert(int pos,int len);
	int Cut(int pos,int len);

	// I/O
	int Load(char *filename);
	int Save(char *filename);

	SetStub();
	void InsertNum(int i,int w);
	void InsertPtr(int i,int p);
	int SaveCom(char *filename);
	int SaveRepCom(char *filename);

	// test code
	int RepEncode(int n,int m);
	int RepDecode();
	int Analyze();

	// info/optimization
	int RemoveRelocs();
	int HandleImports();
	int Process();
	int Build(DProg *p);


	// data
	byte *Buf;
	int Len;

	// data #2
	byte *Buf2;
	int Len2;

	// repetition coder output
	int *RepHead;
	byte *RepData;
	int nRepHead;
	int nRepData;

	IMAGE_FILE_HEADER *BlockPH;
	IMAGE_OPTIONAL_HEADER32 *BlockOH;
	IMAGE_SECTION_HEADER *BlockSH;

	long ImageBase;
	int nSections;

	int DLL;

	// flags
	int ToPrint; //  ToPrint=0: only error/warning messages allowed
	int ToChange; // ToChange=0: no alterations allowed
	// print flags
	int ToPrintRelocs;
	// change flags
	int ToAlign;
	int ToMergeThunks;
	int ToRemoveRelocs;
};

#define BUFD(nn) *((dword*)(&Buf[nn]))



//
// constructor / destructor
//

DProg::DProg()
{
	nSections=0;
	Buf=NULL;
	Len=0;
	Buf2=NULL;
	Len2=0;
	RepHead=NULL;
	nRepHead=0;
	RepData=NULL;
	nRepData=0;

	DLL=0;
}

DProg::~DProg()
{
	if (Buf) free(Buf);
}


//
// memory routines
//

int DProg::Alloc(int len)
{
	Len=len;
	Buf=(byte*)malloc(len);
	if (!Buf) return 1;
	for (int i=0;i<Len;i++) Buf[i]=0;
	return 0;
}

int DProg::Realloc(int len)
{
	Buf=(byte*)realloc(Buf,len);
	if (!Buf) return 1;
	if (len>Len) for (int i=Len;i<len;i++) Buf[i]=0;
	Len=len;
	return 0;
}

int DProg::Set(byte *buf,int len)
{
	Len=len;
	Buf=(byte*)malloc(len);
	if (!Buf) return 1;
	memcpy(Buf,buf,len);
	return 0;
}

int DProg::Insert(int pos,int len)
{
	if (!Buf) return 1;
	if (pos>Len) return 1;
	if (len==0) return 0;
	Buf=(byte*)realloc(Buf,Len+len);
	if (!Buf) return 1;
	if (pos!=Len) memcpy(Buf+pos+len,Buf+pos,Len-pos);
	Len+=len;
	return 0;
}

int DProg::Cut(int pos,int len)
{
	if (!Buf) return 1;
	if (pos+len>Len) return 1;
	if (len==0) return 0;
	// memmove instead of memcpy so that overlapping be allowed
	memmove(Buf+pos,Buf+pos+len,Len-pos-len);
	Buf=(byte*)realloc(Buf,Len-len);
	if (!Buf) return 1;
	Len-=len;
	return 0;
}


//
// I/O routines
//

// load an EXE from disk to Buf
int DProg::Load(char *filename)
{
	FILE *F;
	F=fopen(filename,"rb");
	if (!F) return 1;
	fseek(F,0,SEEK_END);
	Len=ftell(F);
	if (Len<=0) return 1;
	fseek(F,0,SEEK_SET);
	Buf=(byte*)malloc(Len);
	if (!Buf) return 1;
	fread(Buf,1,Len,F);
	fclose(F);
	return 0;
}

// save the EXE in Buf to disk
int DProg::Save(char *filename)
{
	FILE *F;
	int r;

	if (DLL) {
		PE("can't save dll-s.\n");
		return 1;
	}

	if (!Buf) return 1;
	F=fopen(filename,"wb");
	if (!F) return 1;
	r=fwrite(Buf,1,Len,F);
	if (r<=0) return 1;
	fclose(F);
	return 0;
}



//
// the perfect stub...
//

static unsigned char ashrinkStub[]={
    0x4D,0x5A,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
};

static int ashrinkStubLen=64;


int DProg::SetStub()
{
	int i;

	DLL=0;

	free(Buf);
	Buf=(byte*)malloc(64);
	Len=64;
	for (i=0;i<64;i++) Buf[i]=ashrinkStub[i];
	return 0;
}




//
// the .com builder
//

char ExeStr[]="c:\\e.exe";

#define RepComLen sizeof(RepCom)
static byte RepCom[]={
   0xB0,0x00,0xBB,0x5E,0x01,0xBE,0xFF,0xFF,0xBF,0xFF,0xFF,0x8B,0x17,0x43,0x43,0x88, 
   0xD1,0xF3,0xA4,0x88,0xF1,0xF3,0xAA,0x85,0xD2,0x75,0xF0,0xB4,0x5B,0x31,0xC9,0xBA, 
   0xFF,0xFF,0x52,0x52,0xCD,0x21,0x89,0xC3,0xB4,0x40,0xB9,0xFF,0xFF,0xBA,0xFF,0xFF, 
   0xCD,0x21,0xB4,0x3E,0xCD,0x21,0xB4,0x4A,0xBB,0x00,0x10,0xCD,0x21,0x5A,0x89,0x26, 
   0x32,0x01,0xB8,0x00,0x4B,0xBB,0xFF,0xFF,0xCD,0x21,0x8B,0x26,0x32,0x01,0x8C,0xC8, 
   0x8E,0xD0,0x8E,0xD8,0x5A,0xB4,0x41,0xCD,0x21,0xF4,0xF4,0x72,0xF8,0xC3
};

#define DropLen sizeof(DropCom)
static byte DropCom[]={
   0xB4,0x5B,0x31,0xC9,0xBA,0xFF,0xFF,0x52,0x52,0xCD,0x21,0x89,0xC3,0xB4,0x40,0xB9, 
   0xFF,0xFF,0xBA,0x43,0x01,0xCD,0x21,0xB4,0x3E,0xCD,0x21,0xB4,0x4A,0xBB,0x00,0x10, 
   0xCD,0x21,0x5A,0x89,0x26,0x17,0x01,0xB8,0x00,0x4B,0xBB,0xFF,0xFF,0xCD,0x21,0x8B, 
   0x26,0x17,0x01,0x8C,0xC8,0x8E,0xD0,0x8E,0xD8,0x5A,0xB4,0x41,0xCD,0x21,0xF4,0xF4, 
   0x72,0xF8,0xC3
};

void DProg::InsertNum(int i,int w)
{
	Buf2[i]=w&0xFF;
	Buf2[i+1]=(w>>8)&0xFF;
}

void DProg::InsertPtr(int i,int p)
{
	int w=p+0x100;
	Buf2[i]=w&0xFF;
	Buf2[i+1]=(w>>8)&0xFF;
}

// save the program as a COM file; with repetition encoding
int DProg::SaveRepCom(char *filename)
{
	int i;
	FILE *F;
	int r;

	if (DLL) {
		PE("can't convert DLL to COM.\n");
		return 1;
	}
	
	if (!Buf) return 1;
	if (Len+RepComLen>65536) { // approx.
		PE("can't save COM, input EXE is too large (the COM must be 64 kb or less)\n");
		return 1;
	}

	for (i=0;i<8;i++) Buf[2+i]=ExeStr[i]; // insert exe name
	for (i=0;i<14;i++) Buf[10+i]=0; // clear up the place for the EPB

	// data/fill encode
	// TODO: adaptively find the best n, which is here set to a constant 3
	if (r=RepEncode(3,255)) return r;
	RepDecode();

	// build data
	Buf2=new byte[65536];
	if (!Buf2) return 0;
	for (i=0;i<RepComLen;i++) Buf2[i]=RepCom[i];
	for (i=0;i<nRepHead;i++) Buf2[RepComLen+i]=(byte)RepHead[i];
	for (i=0;i<nRepData;i++) Buf2[RepComLen+nRepHead+i]=RepData[i];
	Len2=RepComLen+nRepHead+nRepData;
	InsertPtr(0x003,RepComLen); // dfhead
	InsertPtr(0x006,RepComLen+nRepHead); // dfdata	
	InsertPtr(0x009,RepComLen+nRepHead+nRepData); // exedata
	InsertPtr(0x020,RepComLen+nRepHead+nRepData+2); // filename
	InsertNum(0x02B,Len); // exesize
	InsertPtr(0x02E,RepComLen+nRepHead+nRepData); // exedata
	InsertPtr(0x046,RepComLen+nRepHead+nRepData+10); // exec param block

	// save file
	F=fopen(filename,"wb");
	if (!F) return 1;
	r=fwrite(Buf2,1,Len2,F);
	fclose(F);
	delete Buf2;
	return 0;
}

// save the program as a COM file
int DProg::SaveCom(char *filename)
{
	int i;
	FILE *F;
	int r;
	
	if (DLL) {
		PE("can't convert DLL to COM.\n");
		return 1;
	}
	
	if (!Buf) return 1;
	if (Len+DropLen>65536) {
		PE("can't save COM, input EXE is too large (the COM must be 64 kb or less)\n");
		return 1;
	}

	// build data
	Buf2=new byte[65536];
	if (!Buf2) return 0;
	for (i=0;i<DropLen;i++) Buf2[i]=DropCom[i];
	for (i=0;i<Len;i++) Buf2[DropLen+i]=Buf[i];
	for (i=0;i<8;i++) Buf2[DropLen+2+i]=ExeStr[i]; // insert exe name
	for (i=0;i<14;i++) Buf2[DropLen+10+i]=0; // clear up the place for the EPB
	Len2=DropLen+Len;
	InsertPtr(0x005,DropLen+2); // filename
	InsertNum(0x010,Len); // exesize
//	InsertPtr(0x013,DropLen); // exedata
	InsertPtr(0x02B,DropLen+10); // exec param block

	// save file
	F=fopen(filename,"wb");
	if (!F) return 1;
	r=fwrite(Buf2,1,Len2,F);
	fclose(F);
	delete Buf2;
	return 0;
}



//
// test code
//


#define StorePair {RepHead[nRepHead]=CurData;RepHead[nRepHead+1]=CurFill;CurData=0;CurFill=0;nRepHead+=2;}

// repetition encoder
// "fill"= bytes with value 0; "data"= non-0 bytes
// fill is at least <n> bytes, at most <m> bytes
int DProg::RepEncode(int n,int m)
{
	int i,k;
	int RealFill,WastedFill;
	int CurData,CurFill;
	byte b;

	if (DLL) {
		PE("RepEncode: can't change DLL's.\n");
		return 1;
	}
	
	VMSG("RepEncode encoding: min %d, max %d\n",n,m);
	if (n<=0) return 1;
	if (Len<=0) return 1;

	if (RepHead) delete RepHead;
	if (RepData) delete RepData;

	RepHead=new int[20000];
	nRepHead=0;
	RepData=new byte[Len];
	nRepData=0;

	RealFill=WastedFill=0;
	CurData=CurFill=0;
	for (i=0;i<Len;i++) {
		b=Buf[i];

		if (!b) {
			// actual byte is zero
			RealFill++;
			CurFill++;
		} else {
			// actual byte is non-zero

			if ((CurFill>n)||(CurFill+CurData+1>m)) {
				StorePair // CurData may be 0!
			} else if (CurFill>0) { // 0 < CurFill < n
				WastedFill+=CurFill;
				for (k=0;k<CurFill;k++) RepData[nRepData++]=0;
				CurData+=CurFill;
				CurFill=0;
			}

			RepData[nRepData++]=b;
			CurData++;
		}

		// overflow
		if ((CurData==m)||(CurFill==m)) {
			StorePair
		}

		// end of data
		if (i==Len-1) {
			if ((CurData!=0)||(CurFill!=0)) {
				StorePair
			}
			StorePair // terminator: two 0-s
		}

	}
	i=nRepHead+nRepData;
	float f=(float)i/(float)Len;
	VMSG("result: %4.2f %d/%d bytes\n",f,i,Len);
	VMSG("%d/%d fill wasted, %d header, %d data\n",RealFill,WastedFill,nRepHead,nRepData);
	VMSG("\n");

	return 0;
}

// test code
// repetition decoder
int DProg::RepDecode()
{
	int i;
	int a,b;

	if (DLL) {
		PE("RepEncode: can't change DLL's.\n");
		return 1;
	}

	Buf2=new byte[65536];

	byte *dest=Buf2;
	byte *data=RepData;
	int *head=RepHead;
	while (1) {
		a=*head++;
		b=*head++;
		if ((!a)&&(!b)) break;
		for (i=0;i<a;i++) *dest++=*data++;
		for (i=0;i<b;i++) *dest++=0;
	}

	for (i=0;i<Len;i++) {
		if (Buf[i]!=Buf2[i]) PM("rep[%d] error\n",i);
	}

	return 0;
}

int DProg::Analyze()
{
	int i;
	int freq[256];

	if (!Buf) return 1;
	if (Len<=0) return 1;
	VMSG("\n");
	VMSG("Data analysis\n");
	VMSG("%d bytes\n",Len);
	VMSG("\n");

	
	VMSG("frequency analysis:\n");
	for (i=0;i<256;i++) freq[i]=0;
	for (i=0;i<Len;i++) freq[Buf[i]]++;
	for (i=0;i<256;i++) VMSG("[%3d] %8d\n",i,freq[i]);
	VMSG("\n");

	i=Len-freq[0];
	float f=(float)i/(float)Len;
	VMSG("data: %4.2f %d/%d bytes\n",f,i,Len);
	VMSG("\n");

//	RepEncode(1,10000);
	RepEncode(1,255);
	RepEncode(2,255);
	RepEncode(3,255);
	RepEncode(3,257);
	RepEncode(4,255);
	
	return 0;
}




//
// info/optimization
//



// expected to be called from Process() only!
int DProg::RemoveRelocs()
{
	int i,j;
	long RelRVA;
	long RelPA,RelPS;
	IMAGE_SECTION_HEADER *a;

	if (DLL) {
		PE("RepEncode: can't change DLL's.\n");
		return 1;
	}

	PM("removing relocations...\n");

	if (BlockPH->Characteristics&IMAGE_FILE_RELOCS_STRIPPED) {
		PM("relocations have already been removed\n");
		return 0;
	}
	BlockPH->Characteristics|=IMAGE_FILE_RELOCS_STRIPPED;

	RelRVA=BlockOH->DataDirectory[5].VirtualAddress;
	if ((RelRVA==0)||(BlockOH->DataDirectory[5].Size==0)) {
		PM("there are no relocations\n");
		return 0;
	}
	BlockOH->DataDirectory[5].VirtualAddress=0;
	BlockOH->DataDirectory[5].Size=0; // this is the PHYSICAL size


	// check all sections for embedded relocations, find relocation section
	a=BlockSH;
	RelPA=RelPS=0;
	for (i=0;i<nSections;i++,a++) {
		if (a->PointerToRelocations||a->NumberOfRelocations) {
			PM("can not remove embedded relocations\n");
			return 1;
		}
		if (a->VirtualAddress==RelRVA) {
			RelPA=a->PointerToRawData;
			RelPS=a->SizeOfRawData;
			a->PointerToRawData=0;
			a->SizeOfRawData=0;
		}
	}

	if ((RelPA==0)||(RelPS==0)) {
		PM("relocations have already been stripped\n");
		return 0;
	}


	// update all sections
	a=BlockSH;
	for (i=0;i<nSections;i++,a++) {
		// line numbers and embedded relocations are assumed not to be present
		if (a->PointerToRawData>=RelPA) a->PointerToRawData-=RelPA;
	}


	// remove the physical data of the relocation section
	j=Len-RelPS-RelPA; // number of bytes after the relocations in the file
	if (j<0) {
		PE("relcation pointers are wrong\n"
			"PA=0x%08X=%d, size=%d, file size=%d\n",RelPA,RelPA,RelPS,Len);
		return 1;
	}
	Len-=RelPS;
	if (j==0) return 0; // relocations are at the end of the file
	for (i=0;i<j;i++) Buf[RelPA+i]=Buf[RelPA+RelPS+i];

	// Buf[] is not reallocated
	// (then: Block?H pointers would have to be updated)

	return 0;
}


int DProg::HandleImports()
{
	int i,j;
	long ImpRVA,ImpVS;
	long ImpPA,ImpPS;
	IMAGE_SECTION_HEADER *a;

	PM("imports...\n");

	ImpRVA=BlockOH->DataDirectory[1].VirtualAddress;
	ImpVS=BlockOH->DataDirectory[1].Size;
	if ((ImpRVA==0)||(ImpVS==0)) {
		PM("there are no imports\n");
		return 0;
	}
	
	// find imports
	a=BlockSH;
	ImpPA=ImpPS=0;
	for (i=0;i<nSections;i++,a++) {
		if ((ImpRVA>=a->VirtualAddress)&&(ImpRVA<a->VirtualAddress+a->Misc.VirtualSize)) {
			j=ImpRVA-a->VirtualAddress;
			PM("imports in section#%d, offset %d\n",i,j);
			if (j>=a->SizeOfRawData) {
				PE("imports point to the alignment area\n");
				return 1;
			}
			if (j+ImpVS>a->SizeOfRawData) {
				PE("imports exceed the section\n");
				return 1;
			}
			ImpPA=a->PointerToRawData+j;
			ImpPS=ImpVS;
			break;
		}
	}
	

	if ((ImpPA==0)||(ImpPS==0)) {
		PE("imports are present, but could not be found!\n");
		return 1;
	}

	if (ToMergeThunks) PM("merging thunks\n");
	PM("\n");


	IMAGE_IMPORT_DESCRIPTOR *p=(IMAGE_IMPORT_DESCRIPTOR*)(&Buf[ImpPA]);
	long d=-ImpRVA+ImpPA;
	// assumption: all RVAs point into this section
	// -this is not necessarily true!
	// -there should be a check to verify that each address is between
	// 0 and section_size
	// -the best would be to implement the code for the "general case"
	// when RVAs can point to anywhere


	// terminator: a zero-filled import table entry
	// and FirstThunk *must* have a non-zero value in a valid import table entry
	for (;p->FirstThunk;p++) {
		PM("[%s]\n",&Buf[p->Name+d]);
		PM("time/date: %d\n",p->TimeDateStamp); // used by a binder utility
		PM("forwarder: %d\n",p->ForwarderChain); // should be -1
		if (ToChange) p->TimeDateStamp=0; // may be set to 0
		if (ToChange) p->ForwarderChain=0; // may be set to 0 (even if it should be -1 :)
		if (p->OriginalFirstThunk==p->FirstThunk) PM("merged thunks!\n");
		if (!p->OriginalFirstThunk) PM("merged thunks, original thunk is cleared!\n");
		long *r=(long*)&Buf[p->FirstThunk+d];
		long *q;
		if (!p->OriginalFirstThunk) q=r;
		else q=(long*)&Buf[p->OriginalFirstThunk+d];
		while (*q) {
			if ((*q+d)&1) {
				PE("bad EXE; imports must be word-aligned.\n");
				return 1;
			}
			if (!*r) {
				PE("the real thunk is shorter than the name thunk\n");
				return 1;
			}

			PM("%04X %s\n",*((word*)(&Buf[*q+d])), // hint
				&Buf[*q+2+d]); // function name (ASCIIZ string)
			if (ToChange) *((word*)(&Buf[*q+d]))=0; // set hint to 0
			*q++;
			*r++;
		}
		if (*r) {
			PE("the real thunk is longer than the name thunk\n");
			return 1;
		}

		if ((ToMergeThunks)&&(p->OriginalFirstThunk!=0)&&
			(p->OriginalFirstThunk!=p->FirstThunk)) {
			// merge the two thunks (if they have not already been merged)
			// the real thunk is expected to contain real information
			// (the pointers in the code point there)
			// the name thunk may be cleared
			q=(long*)&Buf[p->OriginalFirstThunk+d];
			r=(long*)&Buf[p->FirstThunk+d];
			p->OriginalFirstThunk=NULL; //p->FirstThunk;
			while (*r) {
				*r=*q;
				*q=NULL;
				*q++;
				*r++;
			}
		}

		PM("\n");
	}

	return 0;
}



static DFlag DPH_Machine[]={
	{IMAGE_FILE_MACHINE_UNKNOWN,
	"IMAGE_FILE_MACHINE_UNKNOWN"},
	{IMAGE_FILE_MACHINE_I386,
	"IMAGE_FILE_MACHINE_I386"},
	{IMAGE_FILE_MACHINE_R3000,
	"IMAGE_FILE_MACHINE_R3000"},
	{IMAGE_FILE_MACHINE_R4000,
	"IMAGE_FILE_MACHINE_R4000"},
	{IMAGE_FILE_MACHINE_R10000,
	"IMAGE_FILE_MACHINE_R10000"},
	{IMAGE_FILE_MACHINE_WCEMIPSV2,
	"IMAGE_FILE_MACHINE_WCEMIPSV2"},
	{IMAGE_FILE_MACHINE_ALPHA,
	"IMAGE_FILE_MACHINE_ALPHA"},
	{IMAGE_FILE_MACHINE_POWERPC,
	"IMAGE_FILE_MACHINE_POWERPC"},
	{IMAGE_FILE_MACHINE_SH3,
	"IMAGE_FILE_MACHINE_SH3"},
	{IMAGE_FILE_MACHINE_SH3E,
	"IMAGE_FILE_MACHINE_SH3E"},
	{IMAGE_FILE_MACHINE_SH4,
	"IMAGE_FILE_MACHINE_SH4"},
	{IMAGE_FILE_MACHINE_ARM,
	"IMAGE_FILE_MACHINE_ARM"},
	{IMAGE_FILE_MACHINE_THUMB,
	"IMAGE_FILE_MACHINE_THUMB"},
	{IMAGE_FILE_MACHINE_IA64,
	"IMAGE_FILE_MACHINE_IA64"},
	{IMAGE_FILE_MACHINE_MIPS16,
	"IMAGE_FILE_MACHINE_MIPS16"},
	{IMAGE_FILE_MACHINE_MIPSFPU,
	"IMAGE_FILE_MACHINE_MIPSFPU"},
	{IMAGE_FILE_MACHINE_MIPSFPU16,
	"IMAGE_FILE_MACHINE_MIPSFPU16"},
	{IMAGE_FILE_MACHINE_ALPHA64,
	"IMAGE_FILE_MACHINE_ALPHA64"},
	{IMAGE_FILE_MACHINE_AXP64,
	"IMAGE_FILE_MACHINE_AXP64"},
};

static DBits DPH_Characteristics[]={
	{IMAGE_FILE_RELOCS_STRIPPED,"Relocation info stripped from file."},
	{IMAGE_FILE_EXECUTABLE_IMAGE,"File is executable  (i.e. no unresolved external references)."},
	{IMAGE_FILE_LINE_NUMS_STRIPPED,"Line numbers stripped from file."},
	{IMAGE_FILE_LOCAL_SYMS_STRIPPED,"Local symbols stripped from file."},
	{IMAGE_FILE_AGGRESIVE_WS_TRIM,"Agressively trim working set."},
	{IMAGE_FILE_LARGE_ADDRESS_AWARE,"Program can handle >2gb addresses."},
	{IMAGE_FILE_BYTES_REVERSED_LO,"Bytes of machine word are reversed."},
	{IMAGE_FILE_32BIT_MACHINE,"32 bit word machine."},
	{IMAGE_FILE_DEBUG_STRIPPED,"Debugging info stripped from file (in .DBG file)."},
	{IMAGE_FILE_REMOVABLE_RUN_FROM_SWAP,"If Image is on removable media, copy and run from the swap file."},
	{IMAGE_FILE_NET_RUN_FROM_SWAP,"If Image is on Net, copy and run from the swap file."},
	{IMAGE_FILE_SYSTEM,"System File."},
	{IMAGE_FILE_DLL,"File is a DLL."},
	{IMAGE_FILE_UP_SYSTEM_ONLY,"File should only be run on a UP machine."},
	{IMAGE_FILE_BYTES_REVERSED_HI,"Bytes of machine word are reversed."},
};

static DFlag DOH_Subsystem[]={
	{IMAGE_SUBSYSTEM_UNKNOWN,"Unknown subsystem."},
	{IMAGE_SUBSYSTEM_NATIVE,"Image doesn't require a subsystem."},
	{IMAGE_SUBSYSTEM_WINDOWS_GUI,"Image runs in the Windows GUI subsystem."},
	{IMAGE_SUBSYSTEM_WINDOWS_CUI,"Image runs in the Windows character subsystem."},
	{IMAGE_SUBSYSTEM_OS2_CUI,"image runs in the OS/2 character subsystem."},
	{IMAGE_SUBSYSTEM_POSIX_CUI,"image runs in the Posix character subsystem."},
	{IMAGE_SUBSYSTEM_NATIVE_WINDOWS,"image is a native Win9x driver."},
	{IMAGE_SUBSYSTEM_WINDOWS_CE_GUI,"Image runs in the Windows CE subsystem."},
};

static DFlag DRelocType[]={
	{IMAGE_REL_BASED_ABSOLUTE,"absolute"}, // reference is absolute, no relocation is necessary
	{IMAGE_REL_BASED_HIGH,"high"},
	{IMAGE_REL_BASED_LOW,"low"},
	{IMAGE_REL_BASED_HIGHLOW,"highlow"},
	{IMAGE_REL_BASED_HIGHADJ,"highadjust"},
	{IMAGE_REL_BASED_MIPS_JMPADDR,"MIPS jump address"}, // wow:)
	{IMAGE_REL_BASED_SECTION,"section"},
	{IMAGE_REL_BASED_REL32,"32 bits"},
	{IMAGE_REL_BASED_MIPS_JMPADDR16,"MIPS jump address (16 bit)"}, // wow:))
	{IMAGE_REL_BASED_IA64_IMM64,"IA64 IMM64"}, // I hope these ones won't ever come up...
	{IMAGE_REL_BASED_DIR64,"DIR64"},
	{IMAGE_REL_BASED_HIGH3ADJ,"high3adjust"},
};

static char *DOH_Directories[]={
	"export directory",
	"import directory",
	"resource directory",
	"exception directory",
	"security directory",
	"base relocation table",
	"debug directory",
	"architecture specific data",
	"RVA of Global Pointer",
	"TLS (thread local storage) directory",
	"load configuration",
	"bound import directory in headers",
	"import address table",
	"delay load import description",
	"COM runtime descriptor",
	"(unknown)",
};

static DBits DSH_Characteristics[]={
	{IMAGE_SCN_TYPE_NO_PAD,"TYPE_NO_PAD - Reserved."},
	{IMAGE_SCN_CNT_CODE,"Section contains code."},
	{IMAGE_SCN_CNT_INITIALIZED_DATA,"Section contains initialized data."},
	{IMAGE_SCN_CNT_UNINITIALIZED_DATA,"Section contains uninitialized data."},
	{IMAGE_SCN_LNK_OTHER,"LNK_OTHER - Reserved."},
	{IMAGE_SCN_LNK_INFO,"Section contains comments or some other type of information."},
	{IMAGE_SCN_LNK_REMOVE,"Section contents will not become part of image."},
	{IMAGE_SCN_LNK_COMDAT,"Section contains comdat."},
	{0x00002000,"Reserved."},
	{IMAGE_SCN_NO_DEFER_SPEC_EXC,"Reset speculative exceptions handling bits in the TLB entries for this section."},
	{IMAGE_SCN_GPREL,"Section content can be accessed relative to GP"},
	{IMAGE_SCN_MEM_FARDATA,"MEM_FARDATA"},
	{IMAGE_SCN_MEM_PURGEABLE,"MEM_PURGEABLE"},
	{IMAGE_SCN_MEM_16BIT,"MEM_16BIT"},
	{IMAGE_SCN_MEM_LOCKED,"MEM_LOCKED"},
	{IMAGE_SCN_MEM_PRELOAD,"MEM_PRELOAD"},

	{IMAGE_SCN_ALIGN_1BYTES,"alignment: 1 bytes"},
	{IMAGE_SCN_ALIGN_2BYTES,"alignment: 2 bytes"},
	{IMAGE_SCN_ALIGN_4BYTES,"alignment: 4 bytes"},
	{IMAGE_SCN_ALIGN_8BYTES,"alignment: 8 bytes"},
	{IMAGE_SCN_ALIGN_16BYTES,"alignment: 16 bytes (default)"},
	{IMAGE_SCN_ALIGN_32BYTES,"alignment: 32 bytes"},
	{IMAGE_SCN_ALIGN_64BYTES,"alignment: 64 bytes"},
	{IMAGE_SCN_ALIGN_128BYTES,"alignment: 128 bytes"},
	{IMAGE_SCN_ALIGN_256BYTES,"alignment: 256 bytes"},
	{IMAGE_SCN_ALIGN_512BYTES,"alignment: 512 bytes"},
	{IMAGE_SCN_ALIGN_1024BYTES,"alignment: 1024 bytes"},
	{IMAGE_SCN_ALIGN_2048BYTES,"alignment: 2048 bytes"},
	{IMAGE_SCN_ALIGN_4096BYTES,"alignment: 4096 bytes"},
	{IMAGE_SCN_ALIGN_8192BYTES,"alignment: 8192 bytes"},
	{0x00F00000,"(unused)"},

	{IMAGE_SCN_LNK_NRELOC_OVFL,"Section contains extended relocations."},
	{IMAGE_SCN_MEM_DISCARDABLE,"Section can be discarded."},
	{IMAGE_SCN_MEM_NOT_CACHED,"Section is not cachable."},
	{IMAGE_SCN_MEM_NOT_PAGED,"Section is not pageable."},
	{IMAGE_SCN_MEM_SHARED,"Section is shareable."},
	{IMAGE_SCN_MEM_EXECUTE,"Section is executable."},
	{IMAGE_SCN_MEM_READ,"Section is readable."},
	{IMAGE_SCN_MEM_WRITE,"Section is writeable."},

};


int DProg::Process()
{
	int i,j;
	byte *b;
	int dir[16];
	int r;

	ImageBase=-1;

	PM("\n");
	PM("program info\n");
	PM("------------->\n");

	// dos stub
	b=Buf;
	if ((b[0]!='M')||(b[1]!='Z')) {
		PM("invalid MZ magic: <%c%c>\n",b[0],b[1]);
		return 1;
	}
	i=*((dword*)(b+0x3C));
	PM("DOS stub size: %d (0x%X) bytes\n",i,i);
	if (i<24) {
		PW("the DOS stub is too small - the .com dropper may not work correcly!\n");
		// the dropper overwrites the first few bytes with its own data
	}
	if (ToChange) {
		// aggressive dos stub optimizer: the EXE will hang from under DOS
		for (j=2;j<i;j++) if ((j<0x3C)||(j>0x3F)) Buf[j]=0;

	}
	b+=i;

	// pe magic
	if ((b[0]!='P')||(b[1]!='E')||(b[2]!='\0')||(b[3]!='\0')) {
		PM("invalid PE magic\n");
		return 1;
	}
	b+=4;

	// pe header
	{
		IMAGE_FILE_HEADER *a=BlockPH=(IMAGE_FILE_HEADER*)b;
		b+=sizeof(IMAGE_FILE_HEADER);
		PM("\nPE header:\n");
		PM("machine: "); PFlag(DPH_Machine,a->Machine); PM("\n");
		PM("%d sections\n",a->NumberOfSections);
		PM("time stamp: 0x%08X\n",a->TimeDateStamp);
		PM("%d symbols, ptr=0x%08X\n",a->NumberOfSymbols,a->PointerToSymbolTable);
		PM("optional header: %d (0x%04X) bytes\n",a->SizeOfOptionalHeader,a->SizeOfOptionalHeader);
		PM("characteristics: ");
		PBits(DPH_Characteristics,a->Characteristics);

		if (!(a->Characteristics&IMAGE_FILE_EXECUTABLE_IMAGE)) {
			PE("program is not executable\n");
			return 1;
		}
		if (a->Characteristics&IMAGE_FILE_SYSTEM) {
			PW("this is a system program\n");
			if (ToChange) {
				PM("system flag removed\n");
				a->Characteristics^=IMAGE_FILE_SYSTEM;
			}
		}
		if (a->Characteristics&IMAGE_FILE_DLL) {
			DLL=1;
			if (ToChange) {
				PE("can't process DLLs\n");
				return 1;
			} else {
				VMSG("file is a DLL\n");
				PW("only a limited range of Dropper's features work with DLL's.\n");
			}
		}
		if (a->SizeOfOptionalHeader==0) {
			PE("optional header has zero size\n");
			return 1;
		};
		if (IMAGE_SIZEOF_NT_OPTIONAL32_HEADER!=224) { // safety code
			PE("internal error\n");
			return 1;
		}
		if (a->SizeOfOptionalHeader!=224) {
			PE("optional header has invalid size\n");
			return 1;
		};

		nSections=a->NumberOfSections;
		if (ToChange) {
			a->Machine=IMAGE_FILE_MACHINE_I386;
			a->TimeDateStamp=0;
			if (a->NumberOfSymbols==0) a->PointerToSymbolTable=0;
			a->Characteristics=IMAGE_FILE_EXECUTABLE_IMAGE|IMAGE_FILE_32BIT_MACHINE;
		}
	}

	// optional header
	// (it is actually not optional but necessary
	// in case of a PE-type Windows executable)
	{
		IMAGE_OPTIONAL_HEADER32 *a=BlockOH=(IMAGE_OPTIONAL_HEADER32*)b;
		b+=sizeof(IMAGE_OPTIONAL_HEADER32);
		PM("\noptional header:\n");
		if (a->Magic!=0x010B) {
			// this is supposed to be a 32 bit optional header
			// (magic: 0x010B, size: 224 bytes)
			// other types are not valid Win'9x executables
			PE("invalid magic (0x%04X)\n",a->Magic);
			return 1;
		}
		PM("linker version: %d.%d\n",a->MajorLinkerVersion,a->MinorLinkerVersion);
		PM("code: %d bytes\n",a->SizeOfCode);
		PM("data: %d bytes (initialized)\n",a->SizeOfInitializedData);
		PM("data: %d bytes (uninitialized)\n",a->SizeOfUninitializedData);
		PM("entry point ptr=0x%08X\n",a->AddressOfEntryPoint);
		PM("base of code ptr=0x%08X\n",a->BaseOfCode);
		PM("base of data ptr=0x%08X\n",a->BaseOfData);
		PM("image base=0x%08X\n",a->ImageBase);

		PM("section alignment:\n");
		PM(" - in memory: 0x%08X (%d) bytes\n",a->SectionAlignment,a->SectionAlignment);
		PM(" - in file: 0x%08X (%d) bytes\n",a->FileAlignment,a->FileAlignment);
		PM("OS version %d.%d\n",
			a->MajorOperatingSystemVersion,
			a->MinorOperatingSystemVersion);
		PM("image version %d.%d\n",
			a->MajorImageVersion,
			a->MinorImageVersion);
		PM("subsystem version %d.%d\n",
			a->MajorSubsystemVersion,
			a->MinorSubsystemVersion);
		PM("Win32 version: %d\n",a->Win32VersionValue);
		PM("image %d bytes\n",a->SizeOfImage);
		PM("headers %d bytes\n",a->SizeOfHeaders);
		PM("checksum: 0x%08X\n",a->CheckSum);
		PM("subsystem: "); PFlag(DOH_Subsystem,a->Subsystem); PM("\n");
		PM("DLL characteristics: %d\n",a->DllCharacteristics);
		PM("stack:\n");
		PM(" - reserve: 0x%08X bytes\n",a->SizeOfStackReserve);
		PM(" - commit: 0x%08X bytes\n",a->SizeOfStackCommit);
		PM("heap:\n");
		PM(" - reserve: 0x%08X bytes\n",a->SizeOfHeapReserve);
		PM(" - commit: 0x%08X bytes\n",a->SizeOfHeapCommit);
		PM("LoaderFlags: %d\n",a->LoaderFlags);

		// directories
		// a Windows program needs 16 directories;
		// most of them are usually unused
		PM("%d directories\n",a->NumberOfRvaAndSizes);
		if (IMAGE_NUMBEROF_DIRECTORY_ENTRIES!=16) { // safety code
			PE("only executables with 16 directories are supported\n");
			return 1;
		}
		if (a->NumberOfRvaAndSizes!=16) {
			PE("there must be 16 directories\n");
			return 1;
		}
		j=0; // number of used directories
		for (i=0;i<16;i++) {
			PM("Directory %02d : ",i);
			if ((a->DataDirectory[i].VirtualAddress==0)||
				(a->DataDirectory[i].Size==0)) {
				if (ToChange) {
					a->DataDirectory[i].VirtualAddress=0;
					a->DataDirectory[i].Size=0;
				}
				dir[i]=0;
				PM("empty");
			} else {
				j++;
				dir[i]=1;
				PM("RVA=0x%08X; size: %d bytes",
					a->DataDirectory[i].VirtualAddress,
					a->DataDirectory[i].Size); // this is the Physical size, not the Virtual size
			}
			PM(" - %s\n",DOH_Directories[i]);
		}
		PM("%d directories are used\n",j);
		for (i=0;i<16;i++) {
			if (!dir[i]) continue;
			switch (i) {
			case 0: // exports
				PW("can not handle exports.\n");
				break;
			case 1: // imports
			case 5: // relocations
				break;
			default: 
				PW("can not handle directory %02d - %s\n",i,DOH_Directories[i]);
			};
		}

		if ((!DLL)&&(a->ImageBase!=0x400000)) { // a DLL may have a different image base
			PE("unconventional image base\n");
			return 1;
		}
		ImageBase=a->ImageBase;
		if (a->SectionAlignment!=0x1000) {
			PW("section alignment should be 0x1000\n");
			// to change the section alignment, the pointers in a code section
			// to the other sections should be updated
			// (this might be possible if relocations are present, but
			// it would be very complicated and it would not get much better:
			// the exe would not get smaller, it is only that the program would
			// take less memory to run)
//			if (ToChange) a->SectionAlignment=0x1000;
		}
		if (a->FileAlignment!=0x200) {
			if (a->FileAlignment<0x200) PW("file alignment must be at least 0x200\n");
			if (a->FileAlignment>0x200) PW("file alignment is unoptimal\n");
			// to change the file alignment, the sections in the file should be
			// updated accordingly (this is not that very complicated to do)
//			if (ToChange) a->FileAlignment=0x200,ToAlign=1;
		}


		if (ToChange) {
#if 1
			// code & data sizes may be zeroed, too (afaik)
			a->SizeOfCode=0;
			a->SizeOfInitializedData=0;
			a->SizeOfUninitializedData=0;
#endif
			a->MajorLinkerVersion=0;
			a->MinorLinkerVersion=0;
			a->MajorOperatingSystemVersion=4;
			a->MinorOperatingSystemVersion=0;
			a->MajorImageVersion=0;
			a->MinorImageVersion=0;
			a->MajorSubsystemVersion=4;
			a->MinorSubsystemVersion=0;
			a->Win32VersionValue=0;
			a->CheckSum=0;
			a->Subsystem=IMAGE_SUBSYSTEM_WINDOWS_GUI;//2;
			a->DllCharacteristics=0;
			// stack: at least 4k is necessary (afaik), but we need it anyway
			if (a->SizeOfStackReserve<0x100000) a->SizeOfStackReserve=0x100000;
			if (a->SizeOfStackCommit<0x1000) a->SizeOfStackCommit=0x1000;
			// heap won't be used, but 4k is necessary (afaik)
			if (a->SizeOfHeapReserve<0x100000) a->SizeOfHeapReserve=0x100000;
			if (a->SizeOfHeapCommit<0x1000) a->SizeOfHeapCommit=0x1000;
			a->LoaderFlags=0; // value does not matter (afaik)
		}
	}

	// section headers
	PM("section headers at PA=0x%08X\n",b-Buf);
	PM("%d sections\n\n",nSections);
	BlockSH=(IMAGE_SECTION_HEADER*)b;
	for (int k=0;k<nSections;k++) {
		PM("section #%d\n",k);
		IMAGE_SECTION_HEADER *a=(IMAGE_SECTION_HEADER*)b;
		b+=sizeof(IMAGE_SECTION_HEADER);
		// a->Name is a null-padded ASCII, not an ASCIIZ string
		char ss[9];
		for (i=0;i<8;i++) ss[i]=a->Name[i];
		ss[8]=0;
		PM("name: %s\n",ss);
		PM("virtual data: 0x%08X - %d bytes\n",a->VirtualAddress,a->Misc.VirtualSize); // ptr in memory (RVA)
		PM("raw data:     0x%08X - %d bytes\n",a->PointerToRawData,a->SizeOfRawData); // ptr in file
		PM("relocations:  0x%08X - %d bytes\n",a->PointerToRelocations,a->NumberOfRelocations); // ptr in file
		PM("line numbers: 0x%08X - %d bytes\n",a->PointerToLinenumbers,a->NumberOfLinenumbers); // ptr in file

		if (ToChange) for (i=0;i<8;i++) a->Name[i]=0; // remove section name

		if ((a->PointerToRelocations!=0)||(a->NumberOfRelocations!=0)) {
			PW("can not handle embedded relocations!\n");
		}
		if ((a->PointerToLinenumbers!=0)||(a->NumberOfLinenumbers!=0)) {
			PW("can not handle line numbers!\n");
		}
		// these won't ever come up anyway (afaik)
		// even if an exe has relocations, they are in a separate section.
		// ie. they are given in the relocation directory and this field is left empty.
		
		PM("characteristics: ");
		PBits(DSH_Characteristics,a->Characteristics);

		
		// try to find the directory <j> to which the current section <a> corresponds to
		j=-1;
		for (i=0;i<16;i++) {
			if (!dir[i]) continue;
			if (a->VirtualAddress!=BlockOH->DataDirectory[i].VirtualAddress) continue;
			j=i;
			break;
		}
		
		if (j==-1) {
			PM("section is not a directory\n");
		} else if (j==1) {
			PM("section contains imports\n");
			if (a->SizeOfRawData!=BlockOH->DataDirectory[1].Size) {
				PW("size of imports declared ambiguously\n");
			}
		} else if (j==5) {
			PM("section contains relocations\n");
			if (a->SizeOfRawData!=BlockOH->DataDirectory[5].Size) {
				PW("size of relocations declared ambiguously\n");
			}

			// print relocations
			if (ToPrintRelocs) {
				PM("\nrelocations:\n"); // "based relocations" in winnt.h
				dword PageRVA=*((dword*)(&Buf[a->PointerToRawData]));
				dword RelocSize=*((dword*)(&Buf[a->PointerToRawData+4]));
				PM("PageRVA=%08X\n",PageRVA);// RelocSize == BlockOH->DataDirectory[5].Size (normally...)
				PM("Size=%d\n",RelocSize);
				int RelocNum=(RelocSize-8);
				if (RelocNum&1) {
					// every relocation is 2 bytes
					// ie. they altogether must occupy an even number of bytes
					PE("invalid relocations\n");
					return 1;
				}
				RelocNum>>=1;
				word *p=(word*)(&Buf[a->PointerToRawData+8]);
				for (int i=0;i<RelocNum;i++) {
					int Type=(*p)>>12;
					int Offset=(*p)&0x0FFF; // TODO: it is in fact signed and not unsigned
					PM("Offset=%03X; Type: ",Offset); PFlag(DRelocType,Type); PM("\n");
					p++;
				}
				PM("\n");
			}

		} else {
			// unknown section; dump it
			PM("unknown type of section\n");
			byte ss[41];
			if (a->SizeOfRawData>1000) {
				PM("section is too big to be dumped\n");
			} else {
				j=0;
				long si=0;
				ss[40]='\0';
				for (i=0;i<a->SizeOfRawData;i++) {
					ss[j]=Buf[a->PointerToRawData+i];
					if (ss[j]<0x20) ss[j]=' ';
					j++;
					if (i==a->SizeOfRawData-1) {
						if (j<40) ss[j]='\0';
						j=40;
					}
					if (j==40) {
						j=0;
						PM("%08X [%s]\n",si,ss);
						si+=40;
					}
				}
			}
		}

		PM("\n");
	}


	// imports
	if (r=HandleImports()) return r;


	// relocations
	if (ToRemoveRelocs) if (r=RemoveRelocs()) return r;


	// TODO: optimize alignment

	return 0;
}



// rebuild a new exe
// !!!!! this code is only in test phase, do not use!
int DProg::Build(DProg *p)
{
	int i,j;

	if (!p) return 1;
	Alloc(2*p->Len); // Alloc() also fills it to 0. TODO: 2*p->Len might be to little

	// DOS header (aggressive, program hangs in DOS)
	Buf[0]='M';
	Buf[1]='Z';
	BUFD(0x3C)=0x40;
	// PE magic
	Buf[0x40]='P';
	Buf[0x41]='E';
	Buf[0x42]=0; // TODO: superfluous
	Buf[0x43]=0; // TODO: superfluous

	// PE header
	BlockPH=(IMAGE_FILE_HEADER*)(Buf+0x44);
	BlockPH->Machine=IMAGE_FILE_MACHINE_I386;
	if (p->BlockPH->NumberOfSections!=p->nSections) {
		PE("invalid number of sections\n");
		return 1;
	}
	BlockPH->NumberOfSections=nSections=p->nSections; // TODO
	BlockPH->TimeDateStamp=0;
	if ((p->BlockPH->NumberOfSymbols)||(p->BlockPH->PointerToSymbolTable)) {
		PW("the program has symbols\n");
	}
	BlockPH->PointerToSymbolTable=0;
	BlockPH->NumberOfSymbols=0;
	BlockPH->SizeOfOptionalHeader=IMAGE_SIZEOF_NT_OPTIONAL32_HEADER; // 224 bytes
	BlockPH->Characteristics=IMAGE_FILE_EXECUTABLE_IMAGE|IMAGE_FILE_32BIT_MACHINE;

	// optional header
	BlockOH=(IMAGE_OPTIONAL_HEADER32*)(BlockPH+1);
	BlockOH->Magic=0x010B;
	BlockOH->MajorLinkerVersion=0;
	BlockOH->MinorLinkerVersion=0;
	BlockOH->SizeOfCode=0;
	BlockOH->SizeOfInitializedData=0;
	BlockOH->SizeOfUninitializedData=0;
	BlockOH->AddressOfEntryPoint=p->BlockOH->AddressOfEntryPoint;
	BlockOH->BaseOfCode=p->BlockOH->BaseOfCode; // might be required? (TODO)
	BlockOH->BaseOfData=p->BlockOH->BaseOfData; // might be required? (TODO)
	if ((p->ImageBase!=p->BlockOH->ImageBase)||(p->ImageBase!=0x400000)) {
		PE("wrong image base\n");
		return 1;
	}
	BlockOH->ImageBase=ImageBase=0x400000;
	BlockOH->SectionAlignment=p->BlockOH->SectionAlignment;
	BlockOH->FileAlignment=p->BlockOH->FileAlignment; // TODO
	BlockOH->MajorOperatingSystemVersion=4;
	BlockOH->MinorOperatingSystemVersion=0;
	BlockOH->MajorImageVersion=0;
	BlockOH->MinorImageVersion=0;
	BlockOH->MajorSubsystemVersion=4;
	BlockOH->MinorSubsystemVersion=0;
	BlockOH->Win32VersionValue=0;
	BlockOH->SizeOfImage=p->BlockOH->SizeOfImage; // virtual size (in memory) (TODO)
	BlockOH->SizeOfHeaders=p->BlockOH->SizeOfHeaders; // physical size of everything before the actual section data (TODO)
	BlockOH->CheckSum=0;
	BlockOH->Subsystem=IMAGE_SUBSYSTEM_WINDOWS_GUI;
	BlockOH->DllCharacteristics=0;
	// stack: at least 4k is necessary (afaik)
//	BlockOH->SizeOfStackReserve=0x100000;
//	BlockOH->SizeOfStackCommit=0x1000;
	BlockOH->SizeOfStackReserve=p->BlockOH->SizeOfStackReserve;
	BlockOH->SizeOfStackCommit=p->BlockOH->SizeOfStackCommit;
	// heap won't be used, but 4k is necessary (afaik)
	BlockOH->SizeOfHeapReserve=0x100000;
	BlockOH->SizeOfHeapCommit=0x1000;
	BlockOH->LoaderFlags=0; // value does not matter (afaik)
	BlockOH->NumberOfRvaAndSizes=16;
//	BlockOH->DataDirectory
//	IMAGE_DATA_DIRECTORY

	// directories
	for (i=0;i<16;i++) {
		BlockOH->DataDirectory[i].VirtualAddress=p->BlockOH->DataDirectory[i].VirtualAddress;
		BlockOH->DataDirectory[i].Size=p->BlockOH->DataDirectory[i].Size;
		if ((BlockOH->DataDirectory[i].VirtualAddress==0)||
			(BlockOH->DataDirectory[i].Size==0)) {
			BlockOH->DataDirectory[i].VirtualAddress=0;
			BlockOH->DataDirectory[i].Size=0;
		}
	}

	// section headers
	BlockSH=(IMAGE_SECTION_HEADER*)(BlockOH+1);
	for (i=0;i<nSections;i++) {
//		sprintf((char*)(&BlockSH[i].Name[0]),"S[%04d]",i);
		for (j=0;j<8;j++) BlockSH[i].Name[j]=0;
		BlockSH[i].Misc.VirtualSize=p->BlockSH[i].Misc.VirtualSize;
		BlockSH[i].VirtualAddress=p->BlockSH[i].VirtualAddress;
		BlockSH[i].SizeOfRawData=p->BlockSH[i].SizeOfRawData;
		BlockSH[i].PointerToRawData=p->BlockSH[i].PointerToRawData;
		BlockSH[i].PointerToRelocations=NULL;
		BlockSH[i].NumberOfRelocations=NULL;
		BlockSH[i].PointerToLinenumbers=NULL;
		BlockSH[i].NumberOfLinenumbers=NULL;
		BlockSH[i].Characteristics=p->BlockSH[i].Characteristics; // TODO: change
	}

	// sections

	return 0;
}







//
// command-line frontend code
//

VimCon Con;
DProg *Prog;
DProg *Prog2;


// command line options
int OptI=0;
int OptO=0;
int OptR=0;
int OptP=0;
int OptC=0;
int OptE=0;
int OptS=0;



void header()
{
	VMSG("Dropper v1.2\n");
	VMSG("Copyright (C) 1999-2002 by Gergely Kutenics - gem@mailbox.hu\n\n");
}

void usage()
{
	VMSG(
	"Usage: dropper.exe [/|-][options] in.exe [out.exe]\n"
	"Options:\n"
	"  i  info\n"
	"  o  optimize (incl. removing relocations)\n"
	"  r  remove relocations (without any further optimization)\n"
	"  p  pack com with repetition encoder\n"
	"  c  output com\n"
	"  e  output exe\n"
	"  s  give an optimal 64 byte mz stub as the exe output\n"
	"Example: /oc to save an optimized <in>.com\n"
//	"Options are case sensitive.\n"
//	"Any option can be turned off with a minus sign as a suffix.\n"
//	"Example: /i to print info, /i- not to print info\n"
//	"Default: /oc\n"
	);
}

int main(int argc,char *argv[])
{
	char *inexe,*outcom,*outexe;
	char *t,*d;
	int n;
	int i;

	Con.Print(1);

	if (argc<=1) {
		header();
		usage();
		return 0;
	}

	Con.LogFile("dropper.log");
	Con.Log(1);
	header();


	//
	// process commandline
	//


	// get options
	for (n=1;argv[n][0]=='/'||argv[n][0]=='-';n++) { // n=0 is 'dropper.exe'
		for (i=1;argv[n][i]!='\0';i++) { // i=0 is the '/' or '-'
			char c=argv[n][i];
			int v=1;
			if (argv[n][i+1]=='-') {
				v=0;
				i++;
			}
			switch (c) {
			case 'i': OptI=v; break;
			case 'o': OptO=v; break;
			case 'c': OptC=v; break;
			case 'e': OptE=v; break;
			case 'p': OptP=v; break;
			case 'r': OptR=v; break;
			case 's': OptS=v; break;
			default:
				VMSG("invalid option [%c][%c]%d\n",c,argv[n][i+1],i);
				usage();
				return 0;
			};
		}
	}
	// n is the index of the next parameter to be processed

	// get input exe
	if (argc<=n) {
		VMSG("error: no input file is specified\n");
		usage();
		return 0;
	}
	inexe=strdup(argv[n]);
	if (strlen(inexe)==0) {
		VMSG("error: invalid input file\n");
		usage();
		return 0;
	}
	n++;

	// get output exe
	if (OptE)
	if (argc<=n) {
		outexe="out.exe";
	} else {
		outexe=strdup(argv[n]);
		n++;
	}

	// get output com
	if (OptC) {
		t=strdup(inexe);
		d=strrchr(t,'.');
		if (d) *d='\0';
		outcom=(char*)malloc(strlen(t)+4+1);
		sprintf(outcom,"%s.com",t);
	}


	//
	// process the file
	//

	VMSG("source: %s\n",inexe);
	if (OptE) VMSG("destination (exe): %s\n",outexe);
	if (OptC) VMSG("destination (com): %s\n",outcom);
	if (OptE) {
		// outcom could be checked, too; but it's not quite important
		if (strcmp(inexe,outexe)==0) {
			VMSG("input and output files must be different\n");
			return 0;
		}
	}

	Prog=new DProg;

	if (Prog->Load(inexe)) {
		VMSG("source file <%s> does not exist.\n",inexe);
		return 0;
	}
	if (OptI) {
		Prog->ToPrint=1;
		Prog->ToPrintRelocs=0; // 1 => print relocation info
	} else {
		Prog->ToPrint=0;
		Prog->ToPrintRelocs=0;
	}
	if (OptO) {
		Prog->ToChange=1;
		Prog->ToAlign=1;
		Prog->ToMergeThunks=1;
		Prog->ToRemoveRelocs=1;
	} else {
		Prog->ToChange=0;
		Prog->ToAlign=0;
		Prog->ToMergeThunks=0;
		Prog->ToRemoveRelocs=0;
	}
	if (OptR) Prog->ToRemoveRelocs=1; // remove relocations (do no other optimization unless OptO is also on)
	i=Prog->Process();
//	Prog->Analyze();
	if (i) {
		VMSG("failed to process file\n");
		Con.Log(0);
		return 0;
	} else {
		VMSG("file [%s] processed successfully\n",inexe);
	}

	if (OptE) {
		if (OptS) Prog->SetStub();
		Prog->Save(outexe);
		VMSG("exe [%s] saved successfully\n",outexe);
	}
	
/* TODO: Build() unfinished
	Prog2=new DProg;
	i=Prog2->Build(Prog);
	if (i) {
		VMSG("failed to build file\n");
	} else {
		if (OptE) Prog2->Save(outexe);
		VMSG("done.\n");
	}
*/


	if (OptC) {
		if (OptP) Prog->SaveRepCom(outcom);
		else Prog->SaveCom(outcom);
		VMSG("com [%s] saved successfully\n",outcom);
	}

	Con.Log(0);
	return 1;
}
