// VIM demo engine - console class
// Copyright (c) 1999-2001 by Gergely Kutenics - gem@mailbox.hu

#ifndef __VIM_BASE
#define __VIM_BASE

#include <windows.h>
#include <windowsx.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <math.h>

// type definitions
typedef unsigned char           byte;
typedef unsigned short int      word;
typedef unsigned long int       dword;
typedef signed char             sbyte;
typedef signed short int        integer;
typedef signed long int         sdword;
typedef dword                   dbool;
typedef byte                    bbool;

// boolean names
#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif
#ifndef ON
#define ON 1
#endif
#ifndef OFF
#define OFF 0
#endif

#define members(xxx) (sizeof(xxx)/sizeof(xxx[0]))


//
// SetLogFile()
//  Sets the log file name.
//  Call it BEFORE SetLogging(1) is called!
//
// SetLogging(n)
//  n=1: Enables logging; opens the file set with SetLogFile for logging.
//       If it hasn't been set, then it uses the default log file.
//  n=0: Disables logging; the messages will be sent to stdout.
//       If a log file was opened, it is closed.
//
// Message()
//  Sends a message to the specified output (either a file or stdout).
// 
// Error()
//  Sends an error message, and quits to the OS.
//


// VimCon: general console class to print/log/buffer messages/warnings/errors
class VimCon
{
public:
	VimCon();
	VimCon(const char *log);
	virtual ~VimCon();
	void Init(const char *log);

	// messaging services
	void Message(const char *s,...); // send message
	void Warning(const char *s,...); // send message, output message box
	void Warning2(const char *head,const char*s,...); // same with own message box header
	void Error(const char *s,...); // send message, terminate program

	void Print(bool n); // 1=print messages to standard output
	void Log(bool n); // 1=log messages to file
	void Buffer(bool n); // 1=send messages to internal buffer
	void LogFile(const char *s);
	void BufferSize(long l);

	long MessagePos() { return nMessagePos; }
	long BufferPos() { return nBufferPos; }
	const char *Buffer() { return Buf; }


protected:
	bool nPrint;
	bool nLog;
	bool nBuffer;
	long nBufferSize; // size of the buffer in bytes; (never set this variable manually; it must be at least about 2000 for safety)
	char *nLogFile;
	FILE *nLogHandle;
	char *Buf;
	long nBufferPos; // nBufferPos=strlen(Buf)
	long nMessagePos; // the number of characters sent as message since last reset
	void CheckBuffer();
	long nMessageSize;
	char *Msg; // temporary buffer for messages
};

extern VimCon Con; // has to be defined and initialized in the main module
#define VMSG Con.Message
#define VWRN Con.Warning
#define VERR Con.Error


#endif
